# ------------------------ : new rwpse ----------------
from typing import Union, Any, Optional
import numpy as np
import torch
import torch.nn.functional as F
import torch_geometric as pyg
from torch_geometric.data import Data, HeteroData
from torch_geometric.transforms import BaseTransform
from torch_scatter import scatter, scatter_add, scatter_max

from torch_geometric.graphgym.config import cfg

from torch_geometric.utils import (
    get_laplacian,
    get_self_loop_attr,
    to_scipy_sparse_matrix,
)
import torch_sparse
from torch_sparse import SparseTensor

from torch_geometric.utils import to_networkx
from .qrw import (
    qrw_on_graph_from_pyg_graph3,
)

import multiprocessing
from functools import partial
import networkx as nx


def add_node_attr(
    data: Data, value: Any, attr_name: Optional[str] = None
) -> Data:
    if attr_name is None:
        if "x" in data:
            x = data.x.view(-1, 1) if data.x.dim() == 1 else data.x
            data.x = torch.cat([x, value.to(x.device, x.dtype)], dim=-1)
        else:
            data.x = value
    else:
        data[attr_name] = value

    return data


@torch.no_grad()
def add_full_qrrwp(
    data,
    walk_length=8,
    attr_name_abs="rrwp",  # name: 'rrwp'
    attr_name_rel="rrwp",  # name: ('rrwp_idx', 'rrwp_val')
    add_identity=True,
    spd=False,
    num_walkers=1,
    start_duration=0.1,
    end_duration=2 * np.pi,
    **kwargs,
):
    device = data.edge_index.device
    ind_vec = torch.eye(walk_length, dtype=torch.float, device=device)
    num_nodes = data.num_nodes
    edge_index, edge_weight = data.edge_index, data.edge_weight

    adj = SparseTensor.from_edge_index(
        edge_index,
        edge_weight,
        sparse_sizes=(num_nodes, num_nodes),
    )

    deg = adj.sum(dim=1)

    pe_list = []
    i = 0
    if add_identity:
        pe_list.append(torch.eye(num_nodes, dtype=torch.float))
        i = i + 1

    durations = np.linspace(start_duration, end_duration, walk_length - 1)
    durations = durations.tolist()

    if walk_length > 2:
        G = to_networkx(data)
        A = nx.adjacency_matrix(G).todense()
        pe_list_update = qrw_on_graph_from_pyg_graph3(
            A,
            duration=durations,
            num_walkers=num_walkers,
        )
        # Identity plus the others
        pe_list += pe_list_update

    pe = torch.stack(pe_list, dim=-1)  # n x n x k
    abs_pe = pe.diagonal().transpose(0, 1)  # n x k

    rel_pe = SparseTensor.from_dense(pe, has_value=True)
    rel_pe_row, rel_pe_col, rel_pe_val = rel_pe.coo()
    rel_pe_idx = torch.stack([rel_pe_row, rel_pe_col], dim=0)

    if spd:
        spd_idx = walk_length - torch.arange(walk_length)
        val = (rel_pe_val > 0).type(torch.float) * spd_idx.unsqueeze(0)
        val = torch.argmax(val, dim=-1)
        rel_pe_val = F.one_hot(val, walk_length).type(torch.float)
        abs_pe = torch.zeros_like(abs_pe)

    data = add_node_attr(data, abs_pe, attr_name=attr_name_abs)
    data = add_node_attr(data, rel_pe_idx, attr_name=f"{attr_name_rel}_index")
    data = add_node_attr(data, rel_pe_val, attr_name=f"{attr_name_rel}_val")
    data.log_deg = torch.log(deg + 1)
    data.deg = deg.type(torch.long)

    return data
